"""
https://www.trafiklab.se/api/trafiklab-apis/sl/deviations
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from dataclasses_json import DataClassJsonMixin

from .common import SL_TZ, TransportMode, dt_field


@dataclass(frozen=True)
class Publish(DataClassJsonMixin):
    from_: datetime = dt_field(alias="from", tzinfo=SL_TZ)
    upto: Optional[datetime] = dt_field(default=None, tzinfo=SL_TZ)


@dataclass(frozen=True)
class Priority(DataClassJsonMixin):
    importance_level: int
    influence_level: int
    urgency_level: int


@dataclass(frozen=True)
class MessageVariant(DataClassJsonMixin):
    header: str
    details: str
    scope_alias: str
    language: str
    weblink: Optional[str] = None


@dataclass(frozen=True)
class StopPoint(DataClassJsonMixin):
    id: int
    name: str


@dataclass(frozen=True)
class ScopeArea(DataClassJsonMixin):
    id: int
    transport_authority: int
    name: str
    type: str
    stop_points: Optional[List[StopPoint]] = None


@dataclass(frozen=True)
class Line(DataClassJsonMixin):
    id: int
    transport_authority: int
    transport_mode: TransportMode
    designation: Optional[str] = None
    name: Optional[str] = None
    group_of_lines: Optional[str] = None


@dataclass(frozen=True)
class Scope(DataClassJsonMixin):
    """Entities affected by the deviation"""

    stop_areas: Optional[List[ScopeArea]] = None
    lines: Optional[List[Line]] = None


@dataclass(frozen=True)
class Deviation(DataClassJsonMixin):
    """information regarding deviations on SLs transport network."""

    version: int
    created: datetime = dt_field(tzinfo=SL_TZ)
    publish: Publish
    priority: Priority
    message_variants: List[MessageVariant]
    modified: Optional[datetime] = dt_field(default=None, tzinfo=SL_TZ)
    deviation_case_id: Optional[int] = None
    scope: Optional[Scope] = None



import re
import sys
from dataclasses import MISSING, field
from datetime import datetime
from enum import Enum
from typing import Optional, Union
from zoneinfo import ZoneInfo

from dataclasses_json import config

ISO_MSEC_PART = re.compile(r"\.(\d{1,3})\+?")
SL_TZ = ZoneInfo("Europe/Stockholm")


def dt_field(alias: Optional[str] = None, default=MISSING, tzinfo=None):
    """wrap `datetime` field with this so that `dataclasses_json` uses correct format"""

    def _dt_from_str_or_ts(v: Union[str, float]):
        if isinstance(v, str):
            # HACK: python prior to 3.11 does not handle ms part correctly
            # in fromisoformat unless it's 3 digits long
            if sys.version_info < (3, 11):
                if msec := ISO_MSEC_PART.search(v):
                    if len(msec.group(1)) < 3:
                        v = v.replace(msec.group(0), f".{msec.group(1).ljust(3, '0')}+")

            dt = datetime.fromisoformat(v)
        else:
            dt = datetime.fromtimestamp(v)

        # make sure that tzinfo is set if it's not None
        if dt.tzinfo is None and tzinfo is not None:
            dt = dt.replace(tzinfo=tzinfo)

        return dt

    return field(
        default=default,
        metadata=config(
            field_name=alias,
            encoder=datetime.isoformat,
            decoder=_dt_from_str_or_ts,
        ),
    )


class TransportMode(str, Enum):
    BUS = "BUS"
    METRO = "METRO"
    TRAM = "TRAM"
    TRAIN = "TRAIN"
    SHIP = "SHIP"
    FETTRY = "FETTRY"
    TAXI = "TAXI"



import pytest

from tsl.clients.deviations import DeviationsClient
from tsl.clients.transport import TransportClient
from tsl.models.departures import Departure, SiteDepartureResponse
from tsl.models.deviations import Deviation, TransportMode
from tsl.models.sites import Site


@pytest.mark.integration
async def test_deviations():
    cl = DeviationsClient()
    deviations = await cl.get_deviations(
        line=["10", "43"],
        transport_mode=[TransportMode.METRO, TransportMode.TRAIN],
    )

    # serialization loop
    raw = Deviation.schema().dumps(deviations, many=True)
    Deviation.schema().loads(raw, many=True)


@pytest.mark.integration
async def test_transport_departures():
    cl = TransportClient()
    response = await cl.get_site_departures(1002)

    assert isinstance(response, SiteDepartureResponse)

    if response.departures:
        assert isinstance(response.departures[0], Departure)

    # serialization loop
    raw = SiteDepartureResponse.schema().dumps(response)
    SiteDepartureResponse.schema().loads(raw)


@pytest.mark.integration
async def test_transport_sites():
    cl = TransportClient()
    sites = await cl.get_sites()

    # serialization loop
    raw = Site.schema().dumps(sites, many=True)
    Site.schema().loads(raw, many=True)



from typing import List, Optional

import aiohttp

from tsl.models.deviations import Deviation, TransportMode

from .common import AsyncClient, UrlParams

__all__ = ("DeviationsClient",)


class DeviationsClient(AsyncClient):
    """
    client for SL Deviations API
    https://www.trafiklab.se/api/trafiklab-apis/sl/deviations/#/
    """

    @staticmethod
    def get_request_url_params(
        future: Optional[bool] = None,
        site: Optional[List[int]] = None,
        line: Optional[List[str]] = None,
        transport_authority: Optional[int] = None,
        transport_mode: Optional[List[TransportMode]] = None,
    ) -> UrlParams:
        """returns url and params to request deviations"""

        params: list[tuple[str, str]] = []
        if future is not None:
            params.append(("future", "true" if future else "false"))
        if site is not None:
            params.extend(("site", str(x)) for x in site)
        if line is not None:
            params.extend(("line", str(x)) for x in line)
        if transport_authority is not None:
            params.append(("transport_authority", str(transport_authority)))
        if transport_mode is not None:
            params.extend(("transport_mode", x.value) for x in transport_mode)

        return UrlParams("https://deviations.integration.sl.se/v1/messages", params)

    async def get_deviations(
        self,
        future: Optional[bool] = None,
        site: Optional[List[int]] = None,
        line: Optional[List[str]] = None,
        transport_authority: Optional[int] = None,
        transport_mode: Optional[List[TransportMode]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> List[Deviation]:

        args = self.get_request_url_params(
            future, site, line, transport_authority, transport_mode
        )

        if session:
            response = await self._request_json(session, args)
        else:
            async with self.session as new_session:
                response = await self._request_json(new_session, args)

        return Deviation.schema().load(response, many=True)



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from dataclasses_json import DataClassJsonMixin

from .common import SL_TZ, dt_field


@dataclass(frozen=True)
class ValidityPeriod(DataClassJsonMixin):
    """The period for which the object is valid"""

    from_: datetime = dt_field(alias="from", tzinfo=SL_TZ)
    to: Optional[datetime] = dt_field(default=None, tzinfo=SL_TZ)


@dataclass(frozen=True)
class Site(DataClassJsonMixin):
    # Unique identifier of a site
    id: int

    # Global unique site identifier
    gid: int

    # The name of the site generaly known to the public
    name: str

    # The period for which the object is valid.
    valid: ValidityPeriod

    # Alias names that describes the same place but with a different name
    alias: Optional[List[str]] = None

    # An abbreviation for the site
    abbreviation: Optional[str] = None

    # Additional note related to the site
    note: Optional[str] = None

    # WGS84 latitude in decimal degrees
    lat: Optional[float] = 0.0

    # WGS84 longitude in decimal degrees
    lon: Optional[float] = 0.0



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import List, Optional

from dataclasses_json import DataClassJsonMixin

from .common import SL_TZ, TransportMode, dt_field


class DepartureState(str, Enum):
    NOTEXPECTED = "NOTEXPECTED"
    NOTCALLED = "NOTCALLED"
    EXPECTED = "EXPECTED"
    CANCELLED = "CANCELLED"
    INHIBITED = "INHIBITED"
    ATSTOP = "ATSTOP"
    BOARDING = "BOARDING"
    BOARDINGCLOSED = "BOARDINGCLOSED"
    DEPARTED = "DEPARTED"
    PASSED = "PASSED"
    MISSED = "MISSED"
    REPLACED = "REPLACED"
    ASSUMEDDEPARTED = "ASSUMEDDEPARTED"


class JourneyState(str, Enum):
    # Do not show departure at all. Some systems might instead indicate that this departure is available only if ordered
    NOTEXPECTED = "NOTEXPECTED"

    # If a not expected dated vehicle journey is never run, it should at some point in time be considered as not run
    NOTRUN = "NOTRUN"

    # Normally show target time for departure
    EXPECTED = "EXPECTED"

    # A symbol or text indicating that the vehicle journey is not yet in progress could be added depending on presentation system configuration
    ASSIGNED = "ASSIGNED"

    # Show departure as cancelled
    CANCELLED = "CANCELLED"

    # If the presentation system only shows vehicles that are in progress, do not show the departure
    SIGNEDON = "SIGNEDON"

    # Normally show target time for departure. A symbol or text indicating that the vehicle journey is at origin, but not yet in progress could be added depending on presentation system configuration. If the presentation system only shows vehicles that are in progress, do not show the departure
    ATORIGIN = "ATORIGIN"

    # Present the current vehicle journey position, i.e. "has left station X Z minutes ago" or "currently at station Y". Systems that cannot present texts of that size should use a symbol or text indicating that the vehicle journey prediction is unreliable
    FASTPROGRESS = "FASTPROGRESS"

    # Present the current vehicle journey position, i.e. "has left station X Z minutes ago" or "currently at station Y"
    NORMALPROGRESS = "NORMALPROGRESS"

    # Present the current vehicle journey position, i.e. "has left station X Z minutes ago" or "currently at station Y" and information that "traffic moves slowly". Systems that cannot present texts of that size should use a symbol or text indicating that the vehicle journey prediction is unreliable
    SLOWPROGRESS = "SLOWPROGRESS"

    # Present the current vehicle journey position, i.e. "has left station X Z minutes ago" or "currently at station Y" and information that there is a "stop in traffic". Systems that cannot present texts of that size should use a symbol or text indicating that the vehicle journey prediction is unreliable
    NOPROGRESS = "NOPROGRESS"

    # If the vehicle system detects that a vehicle is not following the expected route, it can change the state to off route
    OFFROUTE = "OFFROUTE"

    # If the vehicle finally reaches its destination, the vehicle journey receives the state completed, else it is aborted. Once in progress, a cancellation or sign off will be regarded as the monitored vehicle journey has been aborted. If an aborted dated vehicle journey is resumed again, PubTrans will create a new instance of a monitored vehicle journey
    ABORTED = "ABORTED"

    # If the vehicle finally reaches its destination, the vehicle journey receives the state completed, else it is aborted
    COMPLETED = "COMPLETED"

    # If an expected vehicle journey is not cancelled and never becomes in progress, it should at some point in time be considered as assumed completed
    ASSUMEDCOMPLETED = "ASSUMEDCOMPLETED"


class JourneyPredictionState(str, Enum):
    NORMAL = "NORMAL"
    LOSTCONTACT = "LOSTCONTACT"
    UNRELIABLE = "UNRELIABLE"


class JourneyPassengerLevel(str, Enum):
    EMPTY = "EMPTY"
    SEATSAVAILABLE = "SEATSAVAILABLE"
    STANDINGPASSENGERS = "STANDINGPASSENGERS"
    PASSENGERSLEFTBEHIND = "PASSENGERSLEFTBEHIND"
    UNKNOWN = "UNKNOWN"


class StopAreaType(str, Enum):
    BUSTERM = "BUSTERM"
    METROSTN = "METROSTN"
    TRAMSTN = "TRAMSTN"
    RAILWSTN = "RAILWSTN"
    SHIPBER = "SHIPBER"
    FERRYBER = "FERRYBER"
    AIRPORT = "AIRPORT"
    TAXITERM = "TAXITERM"
    UNKNOWN = "UNKNOWN"


@dataclass(frozen=True)
class DepartureJourney(DataClassJsonMixin):
    id: int
    state: JourneyState
    prediction_state: Optional[JourneyPredictionState] = None
    passenger_level: Optional[JourneyPassengerLevel] = None


@dataclass(frozen=True)
class StopAreaReference(DataClassJsonMixin):
    id: int
    name: str
    sname: Optional[str] = None
    type: Optional[StopAreaType] = None


@dataclass(frozen=True)
class StopPointReference(DataClassJsonMixin):
    id: int
    name: Optional[str] = None
    designation: Optional[str] = None


@dataclass(frozen=True)
class LineReference(DataClassJsonMixin):
    id: int
    designation: Optional[str] = None
    transport_mode: Optional[TransportMode] = None
    group_of_lines: Optional[str] = None


@dataclass(frozen=True)
class DepartureDeviation(DataClassJsonMixin):
    importance_level: int
    consequence: str
    message: str


@dataclass(frozen=True)
class Departure(DataClassJsonMixin):
    direction: str
    direction_code: int
    state: DepartureState
    display: str
    journey: DepartureJourney
    stop_area: StopAreaReference
    stop_point: StopPointReference
    line: LineReference
    deviations: List[DepartureDeviation]
    scheduled: datetime = dt_field(tzinfo=SL_TZ)
    expected: Optional[datetime] = dt_field(tzinfo=SL_TZ)
    via: Optional[str] = None
    destination: Optional[str] = None


@dataclass(frozen=True)
class DeviationScope(DataClassJsonMixin):
    description: Optional[str] = None
    lines: Optional[List[LineReference]] = None
    stop_areas: Optional[List[StopAreaReference]] = None
    stop_points: Optional[List[StopPointReference]] = None


@dataclass(frozen=True)
class StopDeviation(DataClassJsonMixin):
    id: int
    importance_level: int
    message: str
    scope: DeviationScope


@dataclass(frozen=True)
class SiteDepartureResponse(DataClassJsonMixin):
    departures: List[Departure]
    stop_deviations: List[StopDeviation]



from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple, Union

import aiohttp

SessionFactory = Callable[[], aiohttp.ClientSession]


class UrlParams(NamedTuple):
    url: str
    params: Union[List[Tuple[str, str]], Dict[str, Any], None]


class AsyncClient:
    def __init__(self, get_session: Optional[SessionFactory] = None) -> None:
        self._get_session = get_session

    @property
    def session(self):
        if factory := self._get_session:
            return factory()

        return aiohttp.ClientSession()

    async def _request_json(self, session: aiohttp.ClientSession, args: UrlParams):
        response = await session.get(
            args.url,
            params=args.params,
            headers={"Content-Type": "application/json"},
        )
        response.raise_for_status()
        json = await response.json()
        return json
